import std.stdio;
import std.string;

extern(C) {
	int displayX11(uint, uint, byte*);
	int displayWin32(uint, uint, byte*);
}

/// Holds the RGB values of a color.
struct Color{
	ubyte red;
	ubyte green;
	ubyte blue;
}

/**
	An in memory image that can read and write to a .bmp file, or display to the screen.
	You can edit the individual pixels with the setPixel functions.
*/
class BMP {
  public:
  	/// Creates an empty image with the given width and height, in pixels. Final argument is the bits per pixel - either 8 (for indexed images) or 24 (for truecolor images) are valid options.
	this(uint w, uint h, int bpp = 24){
		assert(bpp == 24 || bpp == 8); 	// fixme

		wid = w;
		hei = h;

		bitsPerPixel = bpp;

		data.length = hei;
		for(int a = 0; a < hei; a++)
			data[a].length = width;
	}

	/// Loads a .bmp file.
	/// Bugs: the .bmp must be saved as a 24 bit bitmap.
	this(string fname){
		load(fname);
	}

	/** Gets the pixel's red, green, and blue components */
	const Color getPixel(uint x, uint y){
		Color c;
		switch(bitsPerPixel){
		  case 24:
			c.red = data[y][x].TrueColor.red;
			c.green = data[y][x].TrueColor.green;
			c.blue = data[y][x].TrueColor.blue;
		  break;
		  case 8:
			c = palette[data[y][x].Palette8];
		  break;
		}

		return c;
	}

	/** Returns the palette index for the pixel at x,y. Only valid on 8 bit per pixel images. */
	const uint getPixelIndex(uint x, uint y){
		assert(bitsPerPixel == 8);

		return data[y][x].Palette8;
	}

	/// Sets the pixel at x,y to the given color. Only valid on 24 bit image.
	void setPixel(uint x, uint y, Color c){
		assert(bitsPerPixel == 24);

		data[y][x].TrueColor.red = c.red;
		data[y][x].TrueColor.green = c.green;
		data[y][x].TrueColor.blue = c.blue;
	}

	/// Sets the pixel at x,y to the given palette index. Only valid on 8 bit image.
	void setPixel(uint x, uint y, uint c){
		assert(bitsPerPixel == 8);
		data[y][x].Palette8 = c;
	}

	/// Gets the RGB value of a given entry in the palette. Only valid on 8 bit image.
	const Color getPaletteEntry(ubyte n){
		assert(bitsPerPixel == 8);
		return palette[n];
	}

	/// Finds the best match in the existing palette for the given color.
	/// NOT YET IMPLEMENTED
	const ubyte getNearestPaletteEntry(Color c){
		// FIXME
		assert(0);

		return 0;
	}

	/// Overwrites the given palette index with the given color. Only valid on 8 bit image.
	void setPaletteEntry(ubyte n, Color c){
		assert(bitsPerPixel < 16);
		palette[n] = c;
	}

	/// Allocates a new palette entry and sets it to the given color. Only valid on 8 bit image.
	void addPaletteEntry(Color c){
		assert(bitsPerPixel < 16);
		if(numColorsUsed >= 256)
			throw new Exception("out of room for colors");

		palette[numColorsUsed] = c;
		numColorsUsed++;
	}	

	/// width of the image, in pixels.
	const uint width(){
		return wid;
	}

	/// The height of the image, in pixels.
	const uint height(){
		return hei;
	}

	/// Replaces the image data with that loaded from a given .bmp file.
	/// BUGS: only supports loading of 24 bit bmp.
	void load(string filename){
	// FIXME: this only supports 8 bpp
		uint offsetToBits;
		ushort sc; uint sc2;
		uint fileSize;
		int x, y;
		int a, b, w; // scratch variables for a loop below

		FILE* fp = fopen(filename, "rb");
		if(fp is null)
			throw new Exception("open file");
		scope(exit) fclose(fp);

			void read4(uint* what)  { fread(what, 4, 1, fp); }
			void read2(ushort* what){ fread(what, 2, 1, fp); }
			void read1(ubyte* what) { *what = fgetc(fp); }

		a = fgetc(fp); // 'B'
		a = fgetc(fp); // 'M'

		read4(&fileSize); // size of file in bytes
		read2(&sc); 	// reserved
		read2(&sc); 	// reserved
		read4(&offsetToBits); // offset to the bitmap data

		read4(&sc2); // size of BITMAPINFOHEADER

		read4(&wid); // width
		read4(&hei); // height

		for(a = 2+ 4+4+4 +4 +8; a < offsetToBits; a++)
			sc2 = fgetc(fp);

		bitsPerPixel = 24; // FIXME

		data.length = hei;
		for(a = 0; a < hei; a++)
			data[a].length = width;

		// And finally the data

		for(y = height - 1; y >= 0; y--){
			b = 0;
			for(x = 0; x < width; x++) {
					read1(&(data[y][x].TrueColor.blue));
					read1(&(data[y][x].TrueColor.green));
					read1(&(data[y][x].TrueColor.red));
					b += 3;
			}
			w = b%4;
			if(w)
			for(a = 0; a < 4 - w; a++)
				sc = fgetc(fp); // pad until divisible by four
		}
	}

	/// Saves the image to a .bmp file.
	const void save(string filename){
		FILE* fp = fopen(filename/*toStringz(filename)*/, "wb");
		if(fp is null)
			throw new Exception("can't open save file");
		scope(exit) fclose(fp);

		void write4(uint what)  { fwrite(&what, 4, 1, fp); }
		void write2(ushort what){ fwrite(&what, 2, 1, fp); }
		void write1(ubyte what) { fputc(what, fp); }

		ushort offsetToBits;
		if(bitsPerPixel == 8)
			offsetToBits = 1078;
		if (bitsPerPixel == 24)
			offsetToBits = 54;

		uint fileSize = offsetToBits;
		if(bitsPerPixel == 8)
			fileSize += height * (width + width%4);
		if(bitsPerPixel == 24)
			fileSize += height * ((width * 3) + (!((width*3)%4) ? 0 : 4-((width*3)%4)));

		write1('B');
		write1('M');

		write4(fileSize); // size of file in bytes
		write2(0); 	// reserved
		write2(0); 	// reserved
		write4(offsetToBits); // offset to the bitmap data

		write4(40); // size of BITMAPINFOHEADER

		write4(width); // width
		write4(height); // height

		write2(1); // planes
		write2(bitsPerPixel); // bpp
		write4(0); // compression
		write4(0); // size of uncompressed
		write4(0); // x pels per meter
		write4(0); // y pels per meter
		write4(0); // colors used
		write4(0); // colors important

		// And here we write the palette
		if(bitsPerPixel <= 8)
			foreach(c; palette[0..(1 << bitsPerPixel)]){
				write1(c.blue);
				write1(c.green);
				write1(c.red);
				write1(0);
			}

		// And finally the data

		foreach_reverse(row; data){
			int b = 0;
			foreach(pixel; row){
				switch(bitsPerPixel){
				  case 8:
					write1(pixel.Palette8);
					b++;
				  break;
				  case 24:
					write1(pixel.TrueColor.blue);
					write1(pixel.TrueColor.green);
					write1(pixel.TrueColor.red);
					b += 3;
				  break;
				}
			}
			int w = b%4;
			if(w)
			for(int a = 0; a < 4-w; a++)
				write1(0); // pad until divisible by four
		}
	}

	/// Pops up a GUI window displaying the image. Blocks until the user closes the window or presses any key in the window.
	/// BUGS: Extremely slow on X11.
	void display() {
		byte[] bytes;
		bytes.length = wid*hei*3;
		uint idx = 0;
		foreach(row; data)
		foreach(pixel; row) {
			if(bitsPerPixel == 8) {
				Color c = palette[pixel.Palette8];
				bytes[idx++] = c.red;
				bytes[idx++] = c.green;
				bytes[idx++] = c.blue;
			} else if (bitsPerPixel == 24) {
				bytes[idx++] = pixel.TrueColor.red;
				bytes[idx++] = pixel.TrueColor.green;
				bytes[idx++] = pixel.TrueColor.blue;
			} else assert(0);
		}

		version(linux)
			displayX11(wid, hei, bytes.ptr);
		version(Windows)
			displayWin32(wid, hei, bytes.ptr);
	}
  private:
	uint wid;
	uint hei;

	int bitsPerPixel;

	int numColorsUsed; // truly private
	Color[256] palette;

	typedef Pixel[] Row;
	union Pixel{
		struct TrueColor_{
			ubyte blue;
			ubyte green;
			ubyte red;
		}
		TrueColor_ TrueColor;

		ubyte Palette8;
	}
	Row[] data; // pixel = data[y][x], thus, data.length = hei, data[*].length = width
}

version(standalone){
	void main(){
		int w = 255;
		int h = 255;
/*
		auto i = new BMP(w,h,24); // making one on the fly
//		i.addPaletteEntry(Color(255,0,0)); // 8bpp

		for(int b = 0; b < h; b++)
		for(int a = 0; a < w; a++)
//			i.setPixel(a,b,0); // 8 bpp
			i.setPixel(a,b,Color(255,a,0)); // truecolor
*/
		auto i = new BMP("lol.bmp"); // loading from file

		i.display(); // display it to the screen

//		i.save("lol.bmp"); // write it to a file
	}
}
